# coding=utf-8

import pytest
from mock import MagicMock

import pandas as pd

from sparkmagic.livyclientlib.exceptions import BadUserDataException
from sparkmagic.livyclientlib.command import Command
import sparkmagic.utils.constants as constants
from sparkmagic.livyclientlib.sendpandasdftosparkcommand import (
    SendPandasDfToSparkCommand,
)


def test_send_to_scala():
    input_variable_name = "input"
    input_variable_value = pd.DataFrame({"A": [1], "B": [2]})
    output_variable_name = "output"
    maxrows = 1
    sparkcommand = SendPandasDfToSparkCommand(
        input_variable_name, input_variable_value, output_variable_name, maxrows
    )
    sparkcommand._scala_command = MagicMock(return_value=MagicMock())
    sparkcommand.to_command(
        constants.SESSION_KIND_SPARK,
        input_variable_name,
        input_variable_value,
        output_variable_name,
    )
    sparkcommand._scala_command.assert_called_with(
        input_variable_name, input_variable_value, output_variable_name
    )


def test_send_to_r():
    input_variable_name = "input"
    input_variable_value = pd.DataFrame({"A": [1], "B": [2]})
    output_variable_name = "output"
    maxrows = 1
    sparkcommand = SendPandasDfToSparkCommand(
        input_variable_name, input_variable_value, output_variable_name, maxrows
    )
    sparkcommand._r_command = MagicMock(return_value=MagicMock())
    sparkcommand.to_command(
        constants.SESSION_KIND_SPARKR,
        input_variable_name,
        input_variable_value,
        output_variable_name,
    )
    sparkcommand._r_command.assert_called_with(
        input_variable_name, input_variable_value, output_variable_name
    )


def test_send_to_python():
    input_variable_name = "input"
    input_variable_value = pd.DataFrame({"A": [1], "B": [2]})
    output_variable_name = "output"
    maxrows = 1
    sparkcommand = SendPandasDfToSparkCommand(
        input_variable_name, input_variable_value, output_variable_name, maxrows
    )
    sparkcommand._pyspark_command = MagicMock(return_value=MagicMock())
    sparkcommand.to_command(
        constants.SESSION_KIND_PYSPARK,
        input_variable_name,
        input_variable_value,
        output_variable_name,
    )
    sparkcommand._pyspark_command.assert_called_with(
        input_variable_name, input_variable_value, output_variable_name
    )


def test_should_create_a_valid_scala_expression():
    input_variable_name = "input"
    input_variable_value = pd.DataFrame({"A": [1], "B": [2]})
    output_variable_name = "output"

    pandas_df_jsonized = """[{"A":1,"B":2}]"""
    expected_scala_code = '''
        val rdd_json_array = spark.sparkContext.makeRDD("""{}""" :: Nil)
        val {} = spark.read.json(rdd_json_array)'''.format(
        pandas_df_jsonized, output_variable_name
    )

    sparkcommand = SendPandasDfToSparkCommand(
        input_variable_name, input_variable_value, output_variable_name, 1
    )
    assert sparkcommand._scala_command(
        input_variable_name, input_variable_value, output_variable_name
    ) == Command(expected_scala_code)


def test_should_create_a_valid_r_expression():
    input_variable_name = "input"
    input_variable_value = pd.DataFrame({"A": [1], "B": [2]})
    output_variable_name = "output"

    pandas_df_jsonized = """[{"A":1,"B":2}]"""
    expected_r_code = """
        fileConn<-file("temporary_pandas_df_sparkmagics.txt")
        writeLines('{}', fileConn)
        close(fileConn)
        {} <- read.json("temporary_pandas_df_sparkmagics.txt")
        {}.persist()
        file.remove("temporary_pandas_df_sparkmagics.txt")""".format(
        pandas_df_jsonized, output_variable_name, output_variable_name
    )

    sparkcommand = SendPandasDfToSparkCommand(
        input_variable_name, input_variable_value, output_variable_name, 1
    )
    assert sparkcommand._r_command(
        input_variable_name, input_variable_value, output_variable_name
    ) == Command(expected_r_code)


def test_should_create_a_valid_python3_expression():
    input_variable_name = "input"
    input_variable_value = pd.DataFrame({"A": [1], "B": [2]})
    output_variable_name = "output"
    pandas_df_jsonized = """[{"A":1,"B":2}]"""

    expected_python3_code = SendPandasDfToSparkCommand._python_decode

    expected_python3_code += """
        json_array = json_loads_byteified('{}')
        rdd_json_array = spark.sparkContext.parallelize(json_array)
        {} = spark.read.json(rdd_json_array)""".format(
        pandas_df_jsonized, output_variable_name
    )

    sparkcommand = SendPandasDfToSparkCommand(
        input_variable_name, input_variable_value, output_variable_name, 1
    )
    assert sparkcommand._pyspark_command(
        input_variable_name, input_variable_value, output_variable_name
    ) == Command(expected_python3_code)


def test_should_create_a_valid_python2_expression():
    input_variable_name = "input"
    input_variable_value = pd.DataFrame({"A": [1], "B": [2]})
    output_variable_name = "output"
    pandas_df_jsonized = """[{"A":1,"B":2}]"""

    expected_python2_code = SendPandasDfToSparkCommand._python_decode

    expected_python2_code += """
        json_array = json_loads_byteified('{}')
        rdd_json_array = spark.sparkContext.parallelize(json_array)
        {} = spark.read.json(rdd_json_array)""".format(
        pandas_df_jsonized, output_variable_name
    )

    sparkcommand = SendPandasDfToSparkCommand(
        input_variable_name, input_variable_value, output_variable_name, 1
    )
    assert sparkcommand._pyspark_command(
        input_variable_name, input_variable_value, output_variable_name
    ) == Command(expected_python2_code)


def test_should_properly_limit_pandas_dataframe():
    input_variable_name = "input"
    max_rows = 1
    input_variable_value = pd.DataFrame({"A": [0, 1, 2, 3, 4], "B": [5, 6, 7, 8, 9]})
    output_variable_name = "output"

    pandas_df_jsonized = (
        """[{"A":0,"B":5}]"""  # notice we expect json to have dropped all but one row
    )
    expected_scala_code = '''
        val rdd_json_array = spark.sparkContext.makeRDD("""{}""" :: Nil)
        val {} = spark.read.json(rdd_json_array)'''.format(
        pandas_df_jsonized, output_variable_name
    )

    sparkcommand = SendPandasDfToSparkCommand(
        input_variable_name, input_variable_value, output_variable_name, max_rows
    )
    assert sparkcommand._scala_command(
        input_variable_name, input_variable_value, output_variable_name
    ) == Command(expected_scala_code)


def test_should_raise_when_input_is_not_pandas_df():
    with pytest.raises(BadUserDataException):
        input_variable_name = "input"
        input_variable_value = "not a pandas dataframe"
        output_variable_name = "output"
        sparkcommand = SendPandasDfToSparkCommand(
            input_variable_name, input_variable_value, output_variable_name, 1
        )
        sparkcommand.to_command(
            "spark",
            input_variable_name,
            input_variable_value,
            output_variable_name,
        )



import asyncio

from unittest.mock import MagicMock, call, patch

from sparkmagic.kernels.wrapperkernel.sparkkernelbase import SparkKernelBase
from sparkmagic.utils.constants import LANG_PYTHON

kernel = None
execute_cell_mock = None
do_shutdown_mock = None
ipython_display = None
code = "some spark code"


class TestSparkKernel(SparkKernelBase):
    def __init__(self, user_code_parser=None):
        kwargs = {"testing": True}
        if user_code_parser is None:
            user_code_parser = MagicMock(return_value=code)

        super().__init__(
            None, None, None, None, None, LANG_PYTHON, user_code_parser, **kwargs
        )


def setup_function():
    global kernel, execute_cell_mock, do_shutdown_mock, ipython_display

    user_code_parser = MagicMock(return_value=code)
    kernel = TestSparkKernel(user_code_parser)

    kernel._execute_cell_for_user = execute_cell_mock = MagicMock(
        return_value={"test": "ing", "a": "b", "status": "ok"}
    )
    kernel._do_shutdown_ipykernel = do_shutdown_mock = MagicMock()
    kernel.ipython_display = ipython_display = MagicMock()


def teardown_function():
    pass


def test_execute_valid_code():
    # Verify that the execution flows through.
    ret = kernel.do_execute(code, False)

    kernel.user_code_parser.get_code_to_run.assert_called_once_with(code)
    assert execute_cell_mock.called_once_with(ret, True)
    assert execute_cell_mock.return_value is ret
    assert kernel._fatal_error is None

    assert execute_cell_mock.called_once_with(code, True)
    assert ipython_display.send_error.call_count == 0


def test_execute_throws_if_fatal_error_happened():
    # Verify that if a fatal error already happened, we don't run the code and show the fatal error instead.
    fatal_error = "Error."
    kernel._fatal_error = fatal_error

    ret = kernel.do_execute(code, False)

    assert execute_cell_mock.return_value is ret
    assert kernel._fatal_error == fatal_error
    assert execute_cell_mock.called_once_with("None", True)
    assert ipython_display.send_error.call_count == 1


def test_execute_alerts_user_if_an_unexpected_error_happens():
    # Verify that developer error shows developer error (the Github link).
    # Because do_execute is so minimal, we'll assume we have a bug in the _repeat_fatal_error method
    kernel._fatal_error = "Something bad happened before"
    kernel._repeat_fatal_error = MagicMock(side_effect=ValueError)

    ret = kernel.do_execute(code, False)
    assert execute_cell_mock.return_value is ret
    assert execute_cell_mock.called_once_with("None", True)
    assert ipython_display.send_error.call_count == 1


def test_execute_throws_if_fatal_error_happens_for_execution():
    # Verify that the kernel sends the error from Python execution's context to the user
    fatal_error = "Error."
    message = '{}\nException details:\n\t"{}"'.format(fatal_error, fatal_error)
    reply_content = dict()
    reply_content["status"] = "error"
    reply_content["evalue"] = fatal_error

    execute_cell_mock.return_value = reply_content

    ret = kernel._execute_cell(
        code, False, shutdown_if_error=True, log_if_error=fatal_error
    )
    assert execute_cell_mock.return_value is ret
    assert kernel._fatal_error == message
    assert execute_cell_mock.called_once_with("None", True)
    assert ipython_display.send_error.call_count == 1


def test_shutdown_cleans_up():
    # No restart
    kernel._execute_cell_for_user = ecfu_m = MagicMock()
    kernel._do_shutdown_ipykernel = dsi_m = MagicMock()

    kernel.do_shutdown(False)

    ecfu_m.assert_called_once_with("%%_do_not_call_delete_session\n ", True, False)
    dsi_m.assert_called_once_with(False)

    # On restart
    kernel._execute_cell_for_user = ecfu_m = MagicMock()
    kernel._do_shutdown_ipykernel = dsi_m = MagicMock()

    kernel.do_shutdown(True)

    ecfu_m.assert_called_once_with("%%_do_not_call_delete_session\n ", True, False)
    dsi_m.assert_called_once_with(True)


def test_register_auto_viz():
    kernel._register_auto_viz()

    assert (
        call(
            "from autovizwidget.widget.utils import display_dataframe\nip = get_ipython()\nip.display_formatter"
            ".ipython_display_formatter.for_type_by_name('pandas.core.frame', 'DataFrame', display_dataframe)",
            True,
            False,
            None,
            False,
        )
        in execute_cell_mock.mock_calls
    )


def test_change_language():
    kernel._change_language()

    assert (
        call(
            "%%_do_not_call_change_language -l {}\n ".format(LANG_PYTHON),
            True,
            False,
            None,
            False,
        )
        in execute_cell_mock.mock_calls
    )


def test_load_magics():
    kernel._load_magics_extension()

    assert (
        call("%load_ext sparkmagic.kernels", True, False, None, False)
        in execute_cell_mock.mock_calls
    )


def test_delete_session():
    kernel._delete_session()

    assert (
        call("%%_do_not_call_delete_session\n ", True, False)
        in execute_cell_mock.mock_calls
    )


def test_execute_cell_for_user_ipykernel4():
    want = {"status": "OK"}
    # Can't use patch decorator because
    # it fails to patch async functions in Python < 3.8
    with patch(
        "ipykernel.ipkernel.IPythonKernel.do_execute",
        new_callable=MagicMock,
        return_value=want,
    ) as mock_ipy_execute:
        got = TestSparkKernel()._execute_cell_for_user(code="1", silent=True)

        assert mock_ipy_execute.called
        assert want == got


def test_execute_cell_for_user_ipykernel5():
    want = {"status": "OK"}
    # Can't use patch decorator because
    # it fails to patch async functions in Python < 3.8
    with patch(
        "ipykernel.ipkernel.IPythonKernel.do_execute",
        new_callable=MagicMock,
    ) as mock_ipy_execute:
        mock_ipy_execute.return_value = asyncio.Future()
        mock_ipy_execute.return_value.set_result(want)

        got = TestSparkKernel()._execute_cell_for_user(code="1", silent=True)

        assert mock_ipy_execute.called
        assert want == got


def test_execute_cell_for_user_ipykernel6():
    want = {"status": "OK"}
    # Can't use patch decorator because
    # it fails to patch async functions in Python < 3.8
    with patch(
        "ipykernel.ipkernel.IPythonKernel.do_execute", return_value=want
    ) as mock_ipy_execute:
        got = TestSparkKernel()._execute_cell_for_user(code="1", silent=True)
        assert mock_ipy_execute.called
        assert want == got



__version__ = "0.21.0"

from sparkmagic.serverextension.handlers import (
    load_jupyter_server_extension,
)  # noqa: #501


def _jupyter_server_extension_paths():
    return [{"module": "sparkmagic"}]


def _jupyter_nbextension_paths():
    return []



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from sparkmagic.livyclientlib.sendtosparkcommand import SendToSparkCommand
from sparkmagic.livyclientlib.command import Command
from sparkmagic.livyclientlib.exceptions import BadUserDataException

import sparkmagic.utils.configuration as conf

import pandas as pd


class SendPandasDfToSparkCommand(SendToSparkCommand):
    # convert unicode to utf8 or pyspark will mark data as corrupted(and deserialize incorrectly)
    _python_decode = """
        import sys
        import json

        if sys.version_info.major == 2:
            def json_loads_byteified(json_text):
                return _byteify(
                    json.loads(json_text, object_hook=_byteify),
                    ignore_dicts=True
                )
        else:
            def json_loads_byteified(json_text):
                return json.loads(json_text)

        def _byteify(data, ignore_dicts = False):
            if isinstance(data, unicode):
                return data.encode('utf-8')
            if isinstance(data, list):
                return [ _byteify(item, ignore_dicts=True) for item in data ]
            if isinstance(data, dict) and not ignore_dicts:
                return {
                    _byteify(key, ignore_dicts=True): _byteify(value, ignore_dicts=True)
                    for key, value in data.iteritems()
                }
            return data
    """

    def __init__(
        self, input_variable_name, input_variable_value, output_variable_name, max_rows
    ):
        super(SendPandasDfToSparkCommand, self).__init__(
            input_variable_name, input_variable_value, output_variable_name
        )
        self.max_rows = max_rows

    def _scala_command(self, input_variable_name, pandas_df, output_variable_name):
        self._assert_input_is_pandas_dataframe(input_variable_name, pandas_df)
        pandas_json = self._get_dataframe_as_json(pandas_df)

        scala_code = '''
        val rdd_json_array = spark.sparkContext.makeRDD("""{}""" :: Nil)
        val {} = spark.read.json(rdd_json_array)'''.format(
            pandas_json, output_variable_name
        )

        return Command(scala_code)

    def _pyspark_command(self, input_variable_name, pandas_df, output_variable_name):
        self._assert_input_is_pandas_dataframe(input_variable_name, pandas_df)

        pyspark_code = self._python_decode

        pandas_json = self._get_dataframe_as_json(pandas_df)

        pyspark_code += """
        json_array = json_loads_byteified('{}')
        rdd_json_array = spark.sparkContext.parallelize(json_array)
        {} = spark.read.json(rdd_json_array)""".format(
            pandas_json, output_variable_name
        )

        return Command(pyspark_code)

    def _r_command(self, input_variable_name, pandas_df, output_variable_name):
        self._assert_input_is_pandas_dataframe(input_variable_name, pandas_df)
        pandas_json = self._get_dataframe_as_json(pandas_df)

        r_code = """
        fileConn<-file("temporary_pandas_df_sparkmagics.txt")
        writeLines('{}', fileConn)
        close(fileConn)
        {} <- read.json("temporary_pandas_df_sparkmagics.txt")
        {}.persist()
        file.remove("temporary_pandas_df_sparkmagics.txt")""".format(
            pandas_json, output_variable_name, output_variable_name
        )

        return Command(r_code)

    def _get_dataframe_as_json(self, pandas_df):
        return pandas_df.head(self.max_rows).to_json(orient="records")

    def _assert_input_is_pandas_dataframe(
        self, input_variable_name, input_variable_value
    ):
        if not isinstance(input_variable_value, pd.DataFrame):
            wrong_type = input_variable_value.__class__.__name__
            raise BadUserDataException(
                "{} is not a Pandas DataFrame! Got {} instead.".format(
                    input_variable_name, wrong_type
                )
            )



from mock import MagicMock
from time import sleep

from sparkmagic.livyclientlib.livysession import _HeartbeatThread


def test_create_thread():
    session = MagicMock()
    refresh_seconds = 1
    retry_seconds = 2
    heartbeat_thread = _HeartbeatThread(session, refresh_seconds, retry_seconds)

    assert heartbeat_thread.livy_session == session
    assert heartbeat_thread.refresh_seconds == refresh_seconds
    assert heartbeat_thread.retry_seconds == retry_seconds


def test_run_once():
    session = MagicMock()
    refresh_seconds = 0.1
    retry_seconds = 2
    heartbeat_thread = _HeartbeatThread(session, refresh_seconds, retry_seconds, 1)

    heartbeat_thread.start()
    sleep(0.15)
    heartbeat_thread.stop()

    session.refresh_status_and_info.assert_called_once_with()
    assert heartbeat_thread.livy_session is None


def test_run_stops():
    session = MagicMock()
    refresh_seconds = 0.01
    retry_seconds = 2
    heartbeat_thread = _HeartbeatThread(session, refresh_seconds, retry_seconds)

    heartbeat_thread.start()
    sleep(0.1)
    heartbeat_thread.stop()

    assert session.refresh_status_and_info.called
    assert heartbeat_thread.livy_session is None


def test_run_retries():
    msg = "oh noes!"
    session = MagicMock()
    session.refresh_status_and_info = MagicMock(side_effect=ValueError(msg))
    refresh_seconds = 0.1
    retry_seconds = 0.1
    heartbeat_thread = _HeartbeatThread(session, refresh_seconds, retry_seconds, 1)

    heartbeat_thread.start()
    sleep(0.15)
    heartbeat_thread.stop()

    session.refresh_status_and_info.assert_called_once_with()
    session.logger.error.assert_called_once_with(msg)
    assert heartbeat_thread.livy_session is None


def test_run_retries_stops():
    msg = "oh noes!"
    session = MagicMock()
    session.refresh_status_and_info = MagicMock(side_effect=ValueError(msg))
    refresh_seconds = 0.01
    retry_seconds = 0.01
    heartbeat_thread = _HeartbeatThread(session, refresh_seconds, retry_seconds)

    heartbeat_thread.start()
    sleep(0.1)
    heartbeat_thread.stop()

    assert session.refresh_status_and_info.called
    assert session.logger.error.called
    assert heartbeat_thread.livy_session is None



# coding=utf-8
import pytest
from mock import MagicMock

from sparkmagic.livyclientlib.sendstringtosparkcommand import SendStringToSparkCommand
from sparkmagic.livyclientlib.exceptions import BadUserDataException
from sparkmagic.livyclientlib.command import Command
import sparkmagic.utils.constants as constants


def test_send_to_scala():
    input_variable_name = "input"
    input_variable_value = "value"
    output_variable_name = "output"
    sparkcommand = SendStringToSparkCommand(
        input_variable_name, input_variable_value, output_variable_name
    )
    sparkcommand._scala_command = MagicMock(return_value=MagicMock())
    sparkcommand.to_command(
        constants.SESSION_KIND_SPARK,
        input_variable_name,
        input_variable_value,
        output_variable_name,
    )
    sparkcommand._scala_command.assert_called_with(
        input_variable_name, input_variable_value, output_variable_name
    )


def test_send_to_r():
    input_variable_name = "input"
    input_variable_value = "value"
    output_variable_name = "output"
    sparkcommand = SendStringToSparkCommand(
        input_variable_name, input_variable_value, output_variable_name
    )
    sparkcommand._r_command = MagicMock(return_value=MagicMock())
    sparkcommand.to_command(
        constants.SESSION_KIND_SPARKR,
        input_variable_name,
        input_variable_value,
        output_variable_name,
    )
    sparkcommand._r_command.assert_called_with(
        input_variable_name, input_variable_value, output_variable_name
    )


def test_send_to_pyspark():
    input_variable_name = "input"
    input_variable_value = "value"
    output_variable_name = "output"
    sparkcommand = SendStringToSparkCommand(
        input_variable_name, input_variable_value, output_variable_name
    )
    sparkcommand._pyspark_command = MagicMock(return_value=MagicMock())
    sparkcommand.to_command(
        constants.SESSION_KIND_PYSPARK,
        input_variable_name,
        input_variable_value,
        output_variable_name,
    )
    sparkcommand._pyspark_command.assert_called_with(
        input_variable_name, input_variable_value, output_variable_name
    )


def test_to_command_invalid():
    input_variable_name = "input"
    input_variable_value = 42
    output_variable_name = "output"
    sparkcommand = SendStringToSparkCommand(
        input_variable_name, input_variable_value, output_variable_name
    )
    with pytest.raises(
        BadUserDataException,
    ):
        sparkcommand.to_command(
            "invalid",
            input_variable_name,
            input_variable_value,
            output_variable_name,
        )


def test_should_raise_when_input_aint_a_string():
    input_variable_name = "input"
    input_variable_value = 42
    output_variable_name = "output"
    sparkcommand = SendStringToSparkCommand(
        input_variable_name, input_variable_value, output_variable_name
    )
    with pytest.raises(
        BadUserDataException,
    ):
        sparkcommand.to_command(
            "spark",
            input_variable_name,
            input_variable_value,
            output_variable_name,
        )


def test_should_create_a_valid_scala_expression():
    input_variable_name = "input"
    input_variable_value = "value"
    output_variable_name = "output"
    sparkcommand = SendStringToSparkCommand(
        input_variable_name, input_variable_value, output_variable_name
    )
    assert sparkcommand._scala_command(
        input_variable_name, input_variable_value, output_variable_name
    ) == Command('var {} = """{}"""'.format(output_variable_name, input_variable_value))


def test_should_create_a_valid_python_expression():
    input_variable_name = "input"
    input_variable_value = "value"
    output_variable_name = "output"
    sparkcommand = SendStringToSparkCommand(
        input_variable_name, input_variable_value, output_variable_name
    )
    assert sparkcommand._pyspark_command(
        input_variable_name, input_variable_value, output_variable_name
    ) == Command("{} = {}".format(output_variable_name, repr(input_variable_value)))


def test_should_create_a_valid_r_expression():
    input_variable_name = "input"
    input_variable_value = "value"
    output_variable_name = "output"
    sparkcommand = SendStringToSparkCommand(
        input_variable_name, input_variable_value, output_variable_name
    )
    assert sparkcommand._r_command(
        input_variable_name, input_variable_value, output_variable_name
    ) == Command(
        """assign("{}","{}")""".format(output_variable_name, input_variable_value)
    )



# Copyright (c) 2015  aggftw@gmail.com
# Distributed under the terms of the Modified BSD License.
import importlib
from sparkmagic.livyclientlib.endpoint import Endpoint
import sparkmagic.utils.configuration as conf
from sparkmagic.utils.constants import WIDGET_WIDTH
from .abstractmenuwidget import AbstractMenuWidget


class AddEndpointWidget(AbstractMenuWidget):
    def __init__(
        self,
        spark_controller,
        ipywidget_factory,
        ipython_display,
        endpoints,
        endpoints_dropdown_widget,
        refresh_method,
    ):
        # This is nested
        super(AddEndpointWidget, self).__init__(
            spark_controller, ipywidget_factory, ipython_display, True
        )
        self.endpoints = endpoints
        self.endpoints_dropdown_widget = endpoints_dropdown_widget
        self.refresh_method = refresh_method

        # map auth class path string to the instance of the class.
        self.auth_instances = {}
        for auth in conf.authenticators().values():
            module, class_name = (auth).rsplit(".", 1)
            events_handler_module = importlib.import_module(module)
            auth_class = getattr(events_handler_module, class_name)
            self.auth_instances[auth] = auth_class()

        dropdown_options = [(k, v) for k, v in conf.authenticators().items()]
        self.auth_type = self.ipywidget_factory.get_dropdown(
            options=dropdown_options, description="Auth type:"
        )

        # combine all authentication instance's widgets into one list to pass to self.children.
        self.all_widgets = list()
        for _class, instance in self.auth_instances.items():
            for widget in instance.widgets:
                if _class == self.auth_type.value:
                    widget.layout.display = "flex"
                    self.auth = instance
                else:
                    widget.layout.display = "none"
                self.all_widgets.append(widget)

        # Submit widget
        self.submit_widget = self.ipywidget_factory.get_submit_button(
            description="Add endpoint"
        )

        self.auth_type.on_trait_change(self._update_auth)

        self.children = (
            [
                self.ipywidget_factory.get_html(value="<br/>", width=WIDGET_WIDTH),
                self.auth_type,
            ]
            + self.all_widgets
            + [
                self.ipywidget_factory.get_html(value="<br/>", width=WIDGET_WIDTH),
                self.submit_widget,
            ]
        )

        for child in self.children:
            child.parent_widget = self
        self._update_auth()

    def run(self):
        self.auth.update_with_widget_values()
        if self.auth_type.label == "None":
            endpoint = Endpoint(self.auth.url, None)
        else:
            endpoint = Endpoint(self.auth.url, self.auth)
        self.endpoints[self.auth.url] = endpoint
        self.ipython_display.writeln("Added endpoint {}".format(self.auth.url))
        try:
            # We need to call the refresh method because drop down in Tab 2 for endpoints wouldn't
            # refresh with the new value otherwise.
            self.refresh_method()
        except:
            self.endpoints.pop(self.auth.url, None)
            self.refresh_method()
            raise

    def _update_auth(self):
        """Create an instance of the chosen auth type maps to in the config
        file."""
        for widget in self.auth.widgets:
            widget.layout.display = "none"
        self.auth = self.auth_instances.get(self.auth_type.value)
        for widget in self.auth.widgets:
            widget.layout.display = "flex"



import re
from collections import OrderedDict
from enum import Enum
from functools import partial

"""
Creates a HTML table from Spark's Dataframe default string representation

Turns this:
    +---+------+
    | id|animal|
    +---+------+
    |  1|   cat|
    |  2| mouse|
    |  3| horse|
    +---+------+

Into:
    id	animal
    1	cat
    2	mouse
    3	horse
    
    <table>
        <tbody>
            <tr><th>id</th><th>animal</th></tr>
            <tr><td>1</td><td>cat</td>
            </tr><tr><td>2</td><td>mouse</td><
            /tr><tr><td>3</td><td>horse</td></tr>
        </tbody>
    </table>
"""


header_top_pattern = r"[ \t\f\v]*(?P<header_top>\+[-+]*\+)[\n\r]?"
header_content_pattern = r"[ \t\f\v]*(?P<header_content>\|.*\|)[\n\r]"
header_bottom_pattern = r"(?P<header_bottom>[ \t\f\v]*\+[-+]*\+[\n\r])"
row_content_pattern = r"([ \t\f\v]*\|.*\|[\n\r])*"
footer_pattern = r"(?P<footer>[ \t\f\v]*\+[-+]*\+$)"

dataframe_pattern_r = re.compile(
    header_top_pattern
    + header_content_pattern
    + header_bottom_pattern
    + row_content_pattern
    + footer_pattern,
    re.MULTILINE,
)


def extractors(header_top, header_content):
    """Creates functions to pull column values out of Spark DF rows.

    Based on the top of a Dataframe header, identifies start and end index for
    each column value.

    012345678901
    +---+------+
    | id|animal|
    +---+------+
    |  1|   cat|
    |  2| mouse|
    |  3| horse|
    +---+------+

    For example, the `id` column is row[0:4] and `animal` is row[4:11]

    :param header_top The header border top comprise of `+` and `-` marking off
                       demarcating different columns.
                       eg `+---+------+`
    :param header_content The string following the header_top, containing the
                            actual column names
                            eg `| id|animal|`
    :return A dict of column: function that can be applied to string-row
            representation of a a Dataframe, eg `|  1|   cat|`

            In our example:
            {'id': lambda row: row[0:4], 'animal': lambda row: row[4:11]}
    """
    header_pluses = re.finditer(r"\+", header_top)

    def _extract(l, r, row, offset=0):
        return row[offset + l : offset + r].strip()

    def _extractor_iter():
        start = next(header_pluses)
        for end in header_pluses:
            yield partial(_extract, start.end(), end.start())
            start = end

    return OrderedDict((x(header_content), x) for x in _extractor_iter())


class CellComponentType(str, Enum):
    TEXT = "text"
    DF = "DF"


def cell_components_iter(cell):
    """Provides spans for each dataframe in a cell.

    1. Determines if the evaluated output of a cell contains a Spark DF
    2. Splits the cell output on Dataframes
    3. Alternates yielding Plain Text and DF spans of evaluated cell output

    For example if our cell looked like this with provided line numbers:

    0
    1           Random stuff at the start

    45          +---+------+
                | id|animal|
                +---+------+
                |  1|   bat|
                |  2| mouse|
                |  3| horse|
    247         +---+------+

                Random stuff in the middle
    293
    The cell components are
    (CellComponent.TEXT, 0, 45)
    (CellComponentType.DF, 45, 247)
    (CellComponentType.TEXT, 247, 293)
    """
    if not cell:
        return
    df_spans = dataframe_pattern_r.finditer(cell)
    if cell_contains_dataframe(cell):
        df_start, df_end = next(df_spans).span()
        if df_start > 0:
            # Some text before the first Dataframe
            yield (CellComponentType.TEXT, 0, df_start)
        while df_start < len(cell):
            yield (CellComponentType.DF, df_start, df_end)
            try:
                start, end = next(df_spans).span()
                if start > df_end:
                    # Some text before the next Dataframe
                    yield (CellComponentType.TEXT, df_end, start)
                df_start, df_end = start, end
            except StopIteration:
                yield (CellComponentType.TEXT, df_end, len(cell))
                return
    else:
        # Cell does not contain a DF. The whole cell is text.
        yield (CellComponentType.TEXT, 0, len(cell))


def cell_contains_dataframe(cell):
    return dataframe_pattern_r.search(cell) is not None


class CellOutputHtmlParser:
    """Parses cell output and converts it to HTML if applicable."""

    @staticmethod
    def to_html(output):
        return "<br />".join(
            [
                CellOutputHtmlParser.to_html_component(c, output)
                for c in cell_components_iter(output)
            ]
        )

    @staticmethod
    def to_html_component(component_span, cell):
        component_type, start, end = component_span
        if component_type == CellComponentType.DF:
            return DataframeHtmlParser(cell, start, end).to_table()
        elif component_type == CellComponentType.TEXT:
            return "<pre>%s</pre>" % cell[start:end].strip()


class DataframeHtmlParser:
    """Parses a Spark Dataframe and presents it as a HTML table."""

    header_top_r = re.compile(header_top_pattern)
    header_content_r = re.compile(header_content_pattern)

    def __init__(self, cell, start=0, end=None):
        """Creates a Dataframe parser for a single dataframe.

        :param cell The evaluated output of a cell.
                    Cell can contain more than one dataframe, but a single
                    DataframeHtmlParser can only parse table headers/rows for a
                    a single dataframe in the substring cell[start:end]
        """
        self.cell_contents = cell
        end = end or len(self.cell_contents)
        header_spans = DataframeHtmlParser.header_top_r.finditer(
            self.cell_contents, start, end
        )
        parts = {
            "header_top": next(header_spans).span(),
            "header_content": DataframeHtmlParser.header_content_r.search(
                self.cell_contents, start, end
            ).span(),
            "header_bottom": next(header_spans).span(),
            "footer": next(header_spans).span(),
        }
        self.header_content_span = parts["header_content"]
        header_content = self._cell_span(self.header_content_span)

        self.expected_width = len(header_content.strip())

        header_top = self._cell_span(parts["header_top"])
        self.extractors = extractors(header_top.strip(), header_content.strip())
        # The content is between the header-bottom and the footer
        self.content_span = (parts["header_bottom"][1], parts["footer"][0])

    def _cell_span(self, span):
        s, e = span
        return self.cell_contents[s:e]

    def _rowspan_iter(self):
        """Extract each row from the contents of a Dataframe."""
        row_delimiters = re.compile(r"\n").finditer(
            self.cell_contents, self.content_span[0], self.content_span[1]
        )
        start = self.content_span[0]
        for row_delimiter in row_delimiters:
            end, next_start = row_delimiter.span()[0], row_delimiter.span()[1]
            yield (start, end)
            start = next_start

    def row_iter(self, transform=None):
        """Extract and transform each row from a Dataframe.

        Defaults to converting a row to a dict {colName: value}
        """
        _transform = transform or (
            lambda r: {col: x(r) for col, x in self.extractors.items()}
        )
        for rowspan in self._rowspan_iter():
            row = self._cell_span(rowspan).strip()
            if len(row) != self.expected_width:
                raise ValueError(
                    """Expected DF rows to be uniform width (%d)
                                 but found %s (%d)"""
                    % (self.expected_width, row, len(row))
                )
            yield _transform(row)

    def to_table(self):
        """Converts the contents of a notebook cell to a HTML table."""

        header_content = self._cell_span(self.header_content_span)
        table_header_html = self._to_tr(header_content.strip(), is_header=True)

        table_row_iter = self.row_iter(transform=self._to_tr)
        table_body = "".join([r for r in table_row_iter])
        return "<table>%s%s</table>" % (table_header_html, table_body)

    def _to_tr(self, row, is_header=False):
        """Converts a spark dataframe row to a HTML row."""
        tag = "th" if is_header else "td"
        row_content = [x(row) for x in self.extractors.values()]
        row_html = "".join(
            [
                '<%s><pre style="word-break: unset; background-color: unset;">%s</pre></%s>'
                % (tag, rc, tag)
                for rc in row_content
            ]
        )
        return "<tr>%s</tr>" % row_html
